@@ -34,6 +34,7 @@ import gin.tf
 
 import mesh_tensorflow as mtf
 from mesh_tensorflow.transformer import dataset as transformer_dataset
+from mesh_tensorflow.transformer import learning_rate_schedules
 from mesh_tensorflow.transformer import transformer
 import numpy as np
 import pkg_resources
@@ -54,12 +55,17 @@ FLAGS = tf.flags.FLAGS
 _DEFAULT_CONFIG_FILE = "./gin/defaults.gin"
 
 # List of features used by model.
-_INPUT_FEATURES = [
+_MODEL_FEATURES = [
     "inputs", "inputs_position", "inputs_segmentation", "targets",
     "targets_position", "targets_segmentation", "targets_subsegmentation"
 ]
 
 
+def _filter_features(ex):
+  """Filters example features, keeping only valid model features."""
+  return {k: v for k, v in ex.items() if k in _MODEL_FEATURES}
+
+
 def parse_gin_defaults_and_flags():
   """Parses all default gin files and those provided via flags."""
   # Register .gin file search paths with gin
@@ -150,6 +156,12 @@ def build_model(model_type="bitransformer",
      there is no encoder.  Requires a text2self dataset, with targets, but no
      inputs.
 
+  "delimited_lm": an autoregressive language model trained on a text2text
+     dataset.  Each training example is expressed as
+     [<input_tokens>, EOS, <target_tokens>, EOS].  Model checkpoints are
+     compatible with "lm" models.  One strategy is to pretrain as "lm"
+     then fine-tune as "delimited_lm".
+
   "aligned": a non-autoregressive single-stack model (like BERT).  Requires
      a non-text2self dataset with inputs and targets.  The targets and inputs
      have the same length and each entry in the inputs is aligned to the
@@ -167,8 +179,8 @@ def build_model(model_type="bitransformer",
     target: 'bonjour'
 
   Args:
-    model_type: a string, one of "bitransformer", "lm", "aligned", or
-      "bi_teacher_student"
+    model_type: a string, one of "bitransformer", "lm", "delimited_lm",
+      "aligned", or "bi_teacher_student"
     input_vocab_size: an integer
     output_vocab_size: an integer
     layout_rules: optional, input to mtf.convert_to_layout_rules
@@ -188,9 +200,9 @@ def build_model(model_type="bitransformer",
         output_vocab_size=output_vocab_size,
         mesh_shape=mesh_shape,
         layout=layout_rules)
-  elif model_type == "lm" or model_type == "aligned":
+  elif model_type in ["lm", "delimited_lm", "aligned"]:
     return transformer.Unitransformer(
-        autoregressive=model_type == "lm",
+        autoregressive=model_type in ["lm", "delimited_lm"],
         layer_stack=transformer.make_layer_stack(),
         input_vocab_size=input_vocab_size,
         output_vocab_size=output_vocab_size,
@@ -212,17 +224,30 @@ def tpu_mesh_shape(tpu_topology=gin.REQUIRED,
   This function is passed through gin to the argument `mesh_shape` inside the
   function `run`.
 
+  Alternatively, for model_parallelism, pass a mesh_spec (see simd_mesh_impl.py)
+  TODO(noam): describe
+
   Args:
-    tpu_topology: a string - e.g. "2x2"
+    tpu_topology: a string - e.g. "2x2" or "v3-8"
     model_parallelism: an integer - the number of cores per model replica
+      alternatively a list that can be passed to
+      simd_mesh_impl.HierarchicalTiling
     ensemble_parallelism: an optional integer - if present then create an
       "ensemble" mesh-dimension as well, for splitting the models in an
       ensemble.
   Returns:
     a mtf.Shape
   """
-  x, y = tpu_topology.split("x")
-  num_cores = int(x) * int(y) * 2
+  if tpu_topology.startswith("v"):
+    num_cores = int(tpu_topology.split("-")[-1])
+  else:
+    x, y = tpu_topology.split("x")
+    num_cores = int(x) * int(y) * 2
+  if isinstance(model_parallelism, list):
+    # model_parallelism is actually a spec used to
+    # construct a simd_mesh_impl.HierarchicalTiling object
+    return mtf.simd_mesh_impl.HierarchicalTiling.spec_to_mesh_shape(
+        model_parallelism, num_cores)
   data_parallelism = num_cores // model_parallelism
   if ensemble_parallelism:
     data_parallelism //= ensemble_parallelism
@@ -259,16 +284,18 @@ def tpu_estimator_model_fn(model_type,
                            outer_batch_size=1,
                            tpu_summaries=False,
                            predict_fn=None,
+                           score_in_predict_mode=False,
                            variable_filter=None,
                            init_checkpoint=None,
                            ensemble_inputs=None,
                            mesh_devices=None,
-                           model_info_file=None):
+                           model_info_file=None,
+                           hierarchical_tiling_spec=None):
   """Create a TPUEstimator model function.
 
   Args:
-    model_type: a string. One of "bitransformer", "lm", "aligned", or
-      "bi_teacher_student"
+    model_type: a string. One of "bitransformer", "lm", "delimited_lm",
+      "aligned", or "bi_teacher_student"
     transformer_model: a transformer.Unitransformer or transformer.Bitransformer
     vocabulary: a vocabulary.Vocabulary or (inputs_vocabulary,
       targets_vocabulary) tuple. Used for decoding in predict mode.
@@ -283,9 +310,7 @@ def tpu_estimator_model_fn(model_type,
     keep_checkpoint_max: an integer, maximum number of checkpoints to keep
     save_checkpoints_steps: an integer, save a checkpoint every this number of
       steps
-    learning_rate_schedule: an optional function taking the scalar named
-      argument `step` and return the scalar learning rate. Alternatively, a
-      constant.
+    learning_rate_schedule: a constant or a function from step to learning rate
     optimizer: a class extending optimize.Optimizer, required for training
     outer_batch_size: outer batch dimension that could be used to enable the mix
       of data-parallel and model-parallel training of Mixture of Experts (MoE)
@@ -293,6 +318,7 @@ def tpu_estimator_model_fn(model_type,
     tpu_summaries: a boolean, use rewrites to make summaries work on TPU.  This
       may be slow, since it uses a host call hack.
     predict_fn: an optional function, see docs for `run` for more information.
+    score_in_predict_mode: compute log-likelihood scores instead of predictions
     variable_filter: controls which variables are trained.
       If None (default), train all trainable variables.
       If a string regex, train all variables that match this regex.
@@ -309,6 +335,8 @@ def tpu_estimator_model_fn(model_type,
       slice. Only required for GPU.
     model_info_file: an optional string, information about variables and
       operations will be logged to this file during the TRAIN mode.
+    hierarchical_tiling_spec: an optional list that can be passed as the
+      spec argument to simd_mesh_impl.HierarchicalTiling
   Returns:
     a function to be passed to TPUEstimator
   """
@@ -330,6 +358,8 @@ def tpu_estimator_model_fn(model_type,
       a TPUEstimatorSpec
     """
     del labels, config
+    if mode == tf.estimator.ModeKeys.PREDICT and score_in_predict_mode:
+      mode = "score"
     global_step = tf.train.get_global_step()
     if use_tpu and "context" in params:
       ctx = params["context"]
@@ -343,10 +373,18 @@ def tpu_estimator_model_fn(model_type,
       devices_memeory_usage = [worker0_mem] + [0] * (num_hosts - 1)
       var_placer = mtf.utils.BalancedVariablePlacer(device_list,
                                                     devices_memeory_usage)
-      physical_shape = list(
-          params["context"].device_assignment.topology.mesh_shape)
-      logical_to_physical = mtf.simd_mesh_impl.auto_logical_to_physical_tpu(
-          mesh_shape.to_integer_list, physical_shape)
+      physical_shape = [int(i) for i in
+                        params["context"].device_assignment.topology.mesh_shape]
+      physical_shape = (
+          mtf.simd_mesh_impl.physical_shape_3d_from_topology_proto_4d(
+              physical_shape))
+      if hierarchical_tiling_spec is not None:
+        logical_to_physical = mtf.simd_mesh_impl.HierarchicalTiling(
+            hierarchical_tiling_spec,
+            physical_shape).logical_to_physical
+      else:
+        logical_to_physical = mtf.simd_mesh_impl.auto_logical_to_physical_tpu(
+            mesh_shape.to_integer_list, physical_shape)
       mesh_impl = mtf.simd_mesh_impl.SimdMeshImpl(
           mesh_shape, layout_rules, mesh_devices, ctx.device_assignment,
           logical_to_physical=logical_to_physical)
@@ -358,10 +396,19 @@ def tpu_estimator_model_fn(model_type,
     graph = mtf.Graph()
     mesh = mtf.Mesh(graph, "my_mesh", var_placer)
 
-    mtf_features = {}
-    for key, x in features.items():
+    if (outer_batch_size and
+        mode not in [tf.estimator.ModeKeys.PREDICT, "score"]):
       outer_batch_dim = mtf.Dimension("outer_batch", outer_batch_size)
       batch_dim = mtf.Dimension("batch", batch_size // outer_batch_size)
+      batch_dims = [outer_batch_dim, batch_dim]
+    else:
+      batch_dim = mtf.Dimension("batch", batch_size)
+      batch_dims = [batch_dim]
+    ensemble_dims = ([mtf.Dimension("ensemble", ensemble_inputs)]
+                     if ensemble_inputs else [])
+
+    mtf_features = {}
+    for key, x in features.items():
       # Some auxiliary features may have been generated in packing.
       # The names of these new features are of the form
       #   "<original_feature_name>_<suffix>", e.g. "inputs_segmentation".
@@ -369,10 +416,8 @@ def tpu_estimator_model_fn(model_type,
       #   the "_<suffix>".
       feature_length = sequence_length[key.split("_")[0]]
       length_dim = mtf.Dimension("length", feature_length)
-      ensemble_dims = ([mtf.Dimension("ensemble", ensemble_inputs)]
-                       if ensemble_inputs else [])
       feature_shape = mtf.Shape(
-          ensemble_dims + [outer_batch_dim, batch_dim, length_dim])
+          ensemble_dims + batch_dims + [length_dim])
       x = tf.cast(features[key], tf.int32)
       x = tf.reshape(x, feature_shape.to_integer_list)
       if not use_tpu:
@@ -381,20 +426,73 @@ def tpu_estimator_model_fn(model_type,
             x, [x], "import feature %s" % key, summarize=1000, first_n=10)
       mtf_features[key] = mtf.import_fully_replicated(
           mesh, x, feature_shape, name=key)
-      if key == "targets":
-        anon_targets = mtf.anonymize(mtf_features[key])
 
+    def _verify_feature_exists(feature_name, should_exist):
+      if should_exist != (feature_name in mtf_features):
+        message = (
+            "mode=%s model_type=%s should%s have feature %s" %
+            (mode, model_type, "" if should_exist else " not", feature_name))
+        if "lm" in model_type:
+          message += (
+              "\nA common mistake is that model_type=\"lm\" should be used "
+              "with tasks that produce inputs and targets, while "
+              "model_type=\"delimited_lm\" should be used with tasks that "
+              "produce targets only.")
+        raise ValueError(message)
+
+    # Verify that the right features exist, and transform them if necessary
     if mode == tf.estimator.ModeKeys.PREDICT:
-      def _feature_shape(key):
-        feature_length = sequence_length[key.split("_")[0]]
-        return mtf.Shape([
-            mtf.Dimension("batch", batch_size),
-            mtf.Dimension("length", feature_length)
-        ])
-      mtf_features = {
-          k: mtf.reshape(v, _feature_shape(k))
-          for k, v in six.iteritems(mtf_features)
+      _verify_feature_exists("inputs", True)
+      # "targets" may or may not exist depending on whether we are doing
+      # evaluation or open-ended inference.
+    else:
+      _verify_feature_exists("targets", True)
+      _verify_feature_exists("inputs", model_type != "lm")
+      if model_type == "delimited_lm":
+        mtf_features = _dynamic_text2self(mtf_features)
+
+    if mode == "score":
+      # compute log-likelihoods per sequence
+      if predict_fn:
+        # predict_fn contains a custom scoring function
+        # this code-path has not been tested
+        scores = predict_fn(
+            model=transformer_model,
+            features=mtf_features,
+            variable_dtype=get_variable_dtype())
+      targets = mtf_features["targets"]
+      if isinstance(transformer_model, transformer.Unitransformer):
+        length_dim = targets.shape.dims[-1]
+        inputs = mtf.shift(mtf_features["targets"], offset=1,
+                           dim=length_dim, wrap=False)
+      elif isinstance(transformer_model,
+                      (transformer.Bitransformer,
+                       transformer.StudentTeacher)):
+        inputs = mtf_features["inputs"]
+        weights = None
+      else:
+        raise ValueError("unrecognized class")
+      logits, _ = transformer_model.call_simple(
+          inputs=inputs,
+          targets=targets,
+          compute_loss=False,
+          mode=mode,
+          variable_dtype=get_variable_dtype())
+      batch_dim, length_dim, vocab_dim = logits.shape.dims
+      cross_entropy = mtf.layers.softmax_cross_entropy_with_logits(
+          logits, mtf_features["targets"], vocab_dim)
+      cross_entropy *= mtf.cast(
+          mtf.not_equal(targets, 0), cross_entropy.dtype)
+      if mode == "delimited_lm":
+        cross_entropy *= mtf.cast(mtf.logical_not(
+            transformer.delimited_lm_inputs_mask(targets)), cross_entropy.dtype)
+      scores = -mtf.reduce_sum(cross_entropy, reduced_dim=length_dim)
+      scores = mtf.anonymize(scores)
+      lowering = mtf.Lowering(graph, {mesh: mesh_impl}, autostack=autostack)
+      predictions = {
+          "scores": lowering.export_to_tf_tensor(scores)
       }
+    elif mode == tf.estimator.ModeKeys.PREDICT:
       inputs = mtf_features["inputs"]
       if predict_fn:
         mtf_samples = predict_fn(
@@ -408,8 +506,9 @@ def tpu_estimator_model_fn(model_type,
         mtf_samples = transformer_model.sample_autoregressive(
             inputs, variable_dtype=get_variable_dtype(),
             remove_partial_sequences=True)
-      elif isinstance(transformer_model,
-                      (transformer.Bitransformer, transformer.StudentTeacher)):
+      elif isinstance(
+          transformer_model,
+          (transformer.Bitransformer, transformer.StudentTeacher)):
         mtf_samples = transformer_model.decode(
             inputs, variable_dtype=get_variable_dtype())
       else:
@@ -420,7 +519,7 @@ def tpu_estimator_model_fn(model_type,
       inputs = clean_decodes(lowering.export_to_tf_tensor(inputs))
       outputs = clean_decodes(lowering.export_to_tf_tensor(mtf_samples))
 
-      # Detokenize in the graph if supported by the vocabulary and accelerator.
+      # Detokenize in the graph if supported by vocabulary and accelerator.
       def _maybe_detokenize(ids, vocab):
         if not use_tpu and hasattr(vocab, "decode_tf"):
           return vocab.decode_tf(ids)
@@ -433,6 +532,7 @@ def tpu_estimator_model_fn(model_type,
           "inputs": inputs,
           "outputs": outputs}
 
+    if mode in ["score", tf.estimator.ModeKeys.PREDICT]:
       # When exporting a model, we need to communicate to TF-Serving that
       # master variables need to be copied to their slave slice variables.
       # Estimator uses a Scaffold's "local_init_op" for this purpose, so we
@@ -475,9 +575,7 @@ def tpu_estimator_model_fn(model_type,
         logits: a mtf.Tensor
         loss: a mtf.Tensor
       """
-      if model_type == "lm":
-        if "inputs" in mtf_features:
-          mtf_features = _dynamic_text2self(mtf_features)
+      if model_type in ["lm", "delimited_lm"]:
         _, _, length_dim = mtf_features["targets"].shape
         inputs = mtf.shift(mtf_features["targets"], offset=1,
                            dim=length_dim, wrap=False)
@@ -653,35 +751,56 @@ def tpu_estimator_model_fn(model_type,
     elif mode == tf.estimator.ModeKeys.EVAL:
       # perplexity eval
       logits, loss = logits_and_loss(mtf_features)
-      anon_logits = mtf.anonymize(logits)
+      # compute cross-entropy while still on TPU to avoid having to outfeed the
+      # logits, which might be big.
+      logits = mtf.cast(logits, tf.float32)
+      vocab_dim = logits.shape.dims[-1]
+      targets = mtf_features["targets"]
+      cross_entropy = mtf.layers.softmax_cross_entropy_with_logits(
+          logits, targets, vocab_dim)
+      anon_cross_entropy = mtf.anonymize(cross_entropy)
+      predictions = mtf.cast(mtf.argmax(logits, vocab_dim), targets.dtype)
+      anon_predictions = mtf.anonymize(predictions)
+      anon_targets = mtf.anonymize(targets)
+      anon_weights = mtf.layers.weights_nonzero(anon_targets, dtype=tf.float32)
+      if model_type == "delimited_lm":
+        anon_weights *= mtf.cast(
+            mtf.logical_not(transformer.delimited_lm_inputs_mask(anon_targets)),
+            dtype=tf.float32)
+
       lowering = mtf.Lowering(graph, {mesh: mesh_impl}, autostack=autostack)
       tf_loss = tf.cast(lowering.export_to_tf_tensor(loss), tf.float32)
       tf_loss = tf.cast(tf_loss, tf.float32)
-      tf_logits = tf.cast(lowering.export_to_tf_tensor(anon_logits), tf.float32)
+      tf_predictions = lowering.export_to_tf_tensor(anon_predictions)
+      tf_cross_entropy = lowering.export_to_tf_tensor(anon_cross_entropy)
 
-      def simple_metrics(logits, labels):
+      def simple_metrics(xent, predictions, labels, weights):
         """Simple metrics for teacher-forced eval."""
-        weights = tf.cast(tf.not_equal(labels, 0), tf.float32)
-        xent = tf.nn.sparse_softmax_cross_entropy_with_logits(
-            labels=labels, logits=logits)
-        predictions = tf.cast(tf.argmax(logits, axis=-1), labels.dtype)
         token_correct = tf.cast(
             tf.equal(predictions, labels), tf.float32) * weights
-        sequence_correct = tf.to_float(tf.equal(
-            tf.reduce_sum(token_correct, -1),
-            tf.reduce_sum(weights, -1)))
-        sequence_weights = tf.to_float(
-            tf.not_equal(tf.reduce_sum(weights, -1), 0))
+        sequence_correct = tf.cast(
+            tf.equal(tf.reduce_sum(token_correct, -1),
+                     tf.reduce_sum(weights, -1)),
+            tf.float32)
+        sequence_weights = tf.cast(
+            tf.not_equal(tf.reduce_sum(weights, -1), 0),
+            tf.float32)
         # the purpose of "mean_label" is as a checksum to ensure that
         # models were evaluated on the same data.
         return {"neg_log_perplexity": tf.metrics.mean(-xent, weights),
                 "token_accuracy": tf.metrics.mean(token_correct, weights),
                 "sequence_accuracy": tf.metrics.mean(
                     sequence_correct, sequence_weights),
-                "mean_label": tf.metrics.mean(tf.cast(labels, tf.float32))}
+                "mean_label": tf.metrics.mean(tf.cast(labels, tf.float32)),
+                "num_eval_tokens": metric_sum(weights, name="num_eval_tokens"),
+                "max_targets_length": metric_max(tf.reduce_sum(
+                    weights, axis=-1), name="max_targets_length"),
+               }
 
       labels = lowering.export_to_tf_tensor(anon_targets)
-      eval_metrics = (simple_metrics, [tf_logits, labels])
+      weights = lowering.export_to_tf_tensor(anon_weights)
+      eval_metrics = (simple_metrics, [
+          tf_cross_entropy, tf_predictions, labels, weights])
       with mtf.utils.outside_all_rewrites():
         restore_hook = mtf.MtfRestoreHook(lowering)
       return tpu_estimator.TPUEstimatorSpec(
@@ -693,9 +812,34 @@ def tpu_estimator_model_fn(model_type,
   return my_model_fn
 
 
+def metric_sum(values, name=None, **kwargs):
+  del kwargs
+  with tf.variable_scope(name, "metric_sum", [values]):
+    accum = tf.get_variable(
+        "accum", shape=[], dtype=tf.float32, trainable=False,
+        collections=[tf.GraphKeys.LOCAL_VARIABLES],
+        initializer=tf.zeros_initializer())
+    update_op = tf.assign_add(accum, tf.reduce_sum(tf.cast(values, tf.float32)))
+    return accum, update_op
+
+
+def metric_max(values, name=None, **kwargs):
+  del kwargs
+  with tf.variable_scope(name, "metric_max", [values]):
+    accum = tf.get_variable(
+        "accum", shape=[], dtype=tf.float32, trainable=False,
+        collections=[tf.GraphKeys.LOCAL_VARIABLES],
+        initializer=tf.zeros_initializer())
+    update_op = tf.assign(
+        accum, tf.maximum(accum, tf.reduce_max(tf.cast(values, tf.float32))))
+    return accum, update_op
+
+
 def _dynamic_text2self(mtf_features):
   """Convert a packed feature dictionary from text2text into text2self.
 
+  This conversion is used when training a "delimited_lm" model.
+
   This allows us to train a text2self model on data that has been tokenized and
   packed in text2text format.
 
@@ -781,7 +925,7 @@ def _dynamic_text2self(mtf_features):
   return mtf_features
 
 
-def get_inputs_from_file(input_filename, ignore_comments=True):
+def get_inputs_from_file(input_filename, ignore_comments=False):
   """Read data from file and strip new lines."""
   inputs = [line.rstrip() for line in tf.io.gfile.GFile(input_filename)]
 
@@ -801,7 +945,7 @@ def encode_inputs(inputs,
                   batch_size,
                   sequence_length,
                   eos_id=1):
-  """Encode inputs.
+  """Encode string inputs for inference/scoring.
 
   Args:
     inputs: list of strings
@@ -835,6 +979,46 @@ def encode_inputs(inputs,
   return all_input_ids
 
 
+def encode_delimited_lm(inputs,
+                        targets,
+                        vocabulary,
+                        batch_size,
+                        sequence_length,
+                        eos_id=1,
+                        include_final_eos=True):
+  """Encode inputs and targets for scoring a delimited langauge model.
+
+  Args:
+    inputs: list of strings
+    targets: list of strings
+    vocabulary: a mtf.transformer.vocabulary.Vocabulary
+    batch_size: an integer
+    sequence_length: an integer (maximum decode length)
+    eos_id: EOS id
+    include_final_eos: a boolean
+
+  Returns:
+    all_ids: encoded inputs
+  """
+  n = len(inputs)
+  all_ids = []
+  for inp, tgt in zip(inputs, targets):
+    input_ids = inputs_vocabulary(vocabulary).encode(inp.strip()) + [eos_id]
+    target_ids = targets_vocabulary(vocabulary).encode(tgt.strip())
+    if include_final_eos:
+      target_ids.append(eos_id)
+    ids = input_ids + target_ids
+    if len(ids) > sequence_length:
+      ids = ids[:sequence_length]
+    else:
+      ids.extend([0] * (sequence_length - len(ids)))
+    all_ids.append(ids)
+  # pad to make an integral number of batches
+  all_ids.extend([all_ids[0]] * (-n % batch_size))
+  all_ids = np.array(all_ids, dtype=np.int32)
+  return all_ids
+
+
 @gin.configurable
 def decode(estimator,
            input_fn,
@@ -874,6 +1058,25 @@ def decode(estimator,
   return decodes
 
 
+@gin.configurable
+def compute_log_likelihoods(estimator,
+                            input_fn,
+                            checkpoint_path=None):
+  """Decode from an input_fn.
+
+  Args:
+    estimator: a TPUEstimator
+    input_fn: function that returns a tf.Dataset
+    checkpoint_path: an optional string
+
+  Returns:
+    list of floats
+  """
+  result_iter = estimator.predict(
+      input_fn, checkpoint_path=checkpoint_path)
+  return [float(f) for f in result_iter]
+
+
 def write_lines_to_file(lines, filename):
   """Write each line to a filename, replacing the file if it exists.
 
@@ -979,6 +1182,175 @@ def clean_decodes(ids, eos_id=1, pad_id=0, length_axis=-1):
   return tf.where_v2(valid_ids, ids, pad_id)
 
 
+def _score_with_estimator(estimator, input_fn, eval_checkpoint_step, model_dir,
+                          scores_filename, num_examples=None):
+  """For each example returned by input_fn, compute log likelihood.
+
+  Args:
+    estimator: a TPUEstimator
+    input_fn: a function that that returns a tf.data.Dataset with examples
+      containing the string field 'targets' and optionally the field 'inputs'
+    eval_checkpoint_step: int, list of ints, or None, see `eval_model`
+      docstring.
+    model_dir: string, estimator model_dir
+    scores_filename: a string (path of file to write scores to)
+    num_examples: int, the total # of examples being scored, None if unknown
+
+  Returns:
+    a list of floats
+  """
+  checkpoint_path, = get_checkpoint_iterator(eval_checkpoint_step, model_dir)
+
+  result_iter = estimator.predict(input_fn, checkpoint_path=checkpoint_path)
+  scores = [m["scores"] for m in result_iter]
+  # Remove any padding examples
+  scores = scores[:num_examples]
+  if scores_filename is not None:
+    write_lines_to_file(["%f" % f for f in scores], scores_filename)
+  return scores
+
+
+@gin.configurable
+def score_from_strings(estimator, vocabulary, model_type, batch_size,
+                       sequence_length, model_dir, eval_checkpoint_step,
+                       inputs=gin.REQUIRED, targets=gin.REQUIRED,
+                       scores_filename=gin.REQUIRED, eos_id=1, score_eos=True):
+  """Compute log likelihoods per example and write to a text file.
+
+  inputs & targets must either be the same length (in lines) or have inputs
+  evenly divide targets N times, where each input has N decodes sequentially
+  in targets.
+
+  The function returns a list of floats represnenting the log-liekelihood of the
+  target given the input.  If `scores_filename` is present, then these are also
+  written out as a text file, one per line.
+
+  Args:
+    estimator: a TPUEstimator
+    vocabulary: a mtf.transformer.vocabulary.Vocabulary
+    model_type: a string
+    batch_size: an integer
+    sequence_length: an integer or a dict from feature-key to integer
+      the (packed) sequence length, e.g. {"inputs": 512, "targets": 128}
+    model_dir: string, estimator model_dir
+    eval_checkpoint_step: int, list of ints, or None, see `eval_model`
+      docstring.
+    inputs: optional - a list of strings (inputs) the same length as targets
+      alternatively, a string filepath for a text file (one string per line)
+    targets: a list of strings (targets)
+      alternatively, a string filepath for a text file (one string per line)
+    scores_filename: a string (path of file to write)
+    eos_id: EOS id
+    score_eos: a boolean - whether to score the final eos token of each line
+      If this is set to false, the scores can be interpreted as prefix
+      log-likelihoods
+  Returns:
+    a list of floats
+  """
+  if isinstance(inputs, str):
+    inputs = get_inputs_from_file(inputs)
+  if isinstance(targets, str):
+    targets = get_inputs_from_file(targets)
+  has_inputs = inputs is not None
+  if has_inputs:
+    if len(inputs) < len(targets):
+      # We assume that the targets file contains n targets for each input.
+      # So we repeat each input n times.
+      if len(targets) % len(inputs):
+        raise ValueError("len(inputs) must divide len(targets), got %d and %d"
+                         % (len(inputs), len(targets)))
+      repeats = len(targets) // len(inputs)
+      inputs = [inputs[i // repeats] for i in range(len(targets))]
+    elif len(targets) < len(inputs):
+      # `targets` is a list of one string.  Use it as a target for all inputs.
+      if len(targets) != 1:
+        raise ValueError("Expected only one target string")
+      targets = targets * len(inputs)
+  if model_type == "delimited_lm":
+    all_target_ids = encode_delimited_lm(
+        inputs,
+        targets,
+        vocabulary,
+        batch_size,
+        sequence_length["targets"],
+        eos_id=eos_id)
+    has_inputs = False
+  else:
+    if has_inputs:
+      all_input_ids = encode_inputs(inputs, vocabulary, model_type, batch_size,
+                                    sequence_length["inputs"], eos_id=eos_id)
+    all_target_ids = encode_inputs(
+        targets, vocabulary, model_type, batch_size,
+        sequence_length["targets"], eos_id=eos_id if score_eos else 0)
+
+  def input_fn(params):
+    del params
+    m = ({"inputs": all_input_ids, "targets": all_target_ids} if has_inputs
+         else {"targets": all_target_ids})
+    dataset = tf.data.Dataset.from_tensor_slices(m)
+    dataset = dataset.flat_map(tf.data.Dataset.from_tensors)
+    dataset = dataset.batch(batch_size, drop_remainder=True)
+    return dataset.prefetch(tf.data.experimental.AUTOTUNE)
+
+  return _score_with_estimator(estimator, input_fn, eval_checkpoint_step,
+                               model_dir, scores_filename, len(targets))
+
+
+@gin.configurable
+def score_from_dataset(estimator, vocabulary, batch_size, sequence_length,
+                       model_dir, eval_checkpoint_step, dataset_split,
+                       score_dataset_fn=None, scores_filename=gin.REQUIRED):
+  """Compute log likelihoods per example and write to a text file.
+
+
+
+  The function returns a list of floats represnenting the log-liekelihood of the
+  target given the input.  If `scores_filename` is present, then these are also
+  written out as a text file, one per line.
+
+  Args:
+    estimator: a TPUEstimator
+    vocabulary: a mtf.transformer.vocabulary.Vocabulary
+    batch_size: an integer
+    sequence_length: an integer or a dict from feature-key to integer
+      the (packed) sequence length, e.g. {"inputs": 512, "targets": 128}
+    model_dir: string, estimator model_dir
+    eval_checkpoint_step: int, list of ints, or None, see `eval_model`
+      docstring.
+        dataset_split: a string
+    score_dataset_fn: A function returning a list of dataset.EvalDataset tuples.
+      See `eval_dataset_fn` argument to `eval_model` for details.
+    scores_filename: a string (path of file to write)
+
+  Returns:
+    a list of floats
+  """
+  scoring_datasets = score_dataset_fn(
+      sequence_length=sequence_length,
+      vocabulary=vocabulary,
+      dataset_split=dataset_split)
+  if len(scoring_datasets) != 1:
+    raise ValueError("Only scoring from a single dataset supported.")
+  scoring_dataset = scoring_datasets[0]
+
+  def input_fn(params):
+    """Eval input function for estimator."""
+    del params
+    dataset = scoring_dataset.dataset_fn()
+    dataset = dataset.map(_filter_features)
+    dataset = dataset.batch(batch_size, drop_remainder=False)
+    # Pad the final batch.
+    dataset = transformer_dataset.trim_and_pad_dataset(
+        dataset, length=batch_size)
+    dataset = dataset.prefetch(tf.data.experimental.AUTOTUNE)
+    return dataset
+
+  # TODO(dei): Since we pass in the num_examples as None, scores for the
+  # padding examples will get written to the output file. Should fix this.
+  return _score_with_estimator(estimator, input_fn, eval_checkpoint_step,
+                               model_dir, scores_filename, None)
+
+
 def get_estimator(model_type, vocabulary, mesh_shape,
                   layout_rules, model_dir, batch_size, sequence_length,
                   autostack, learning_rate_schedule, keep_checkpoint_max,
@@ -1086,7 +1458,8 @@ def get_estimator(model_type, vocabulary, mesh_shape,
 
 
 def train_model(estimator, vocabulary, sequence_length, batch_size,
-                train_dataset_fn, train_steps, ensemble_inputs):
+                train_dataset_fn, train_steps, ensemble_inputs,
+                dataset_split="train"):
   """Train a Mesh-TF model.
 
   Args:
@@ -1107,6 +1480,7 @@ def train_model(estimator, vocabulary, sequence_length, batch_size,
       train an ensemble where each model gets different inputs. You also need to
       configure Unitransformer.ensemble  to the right size. If None, then all
       models are trained on the same inputs.
+    dataset_split: str, which dataset split to train on.
   """
 
   def input_fn(params):
@@ -1114,7 +1488,7 @@ def train_model(estimator, vocabulary, sequence_length, batch_size,
     dataset = train_dataset_fn(
         sequence_length=sequence_length,
         vocabulary=vocabulary,
-        dataset_split="train")
+        dataset_split=dataset_split)
     dataset = dataset.repeat().batch(
         batch_size * (ensemble_inputs or 1), drop_remainder=True)
     dataset = dataset.prefetch(tf.data.experimental.AUTOTUNE)
@@ -1277,9 +1651,7 @@ def eval_model(estimator, vocabulary, sequence_length, batch_size,
       # Only cache targets for those tasks with eval functions provides
       if eval_dataset.metric_fns:
         ds = eval_dataset.dataset_fn()
-        # Only pass those variables which will be used for decoding
-        ds = ds.map(
-            lambda x: {k: v for k, v in x.items() if k in _INPUT_FEATURES})
+        ds = ds.map(_filter_features)
         combined_ds = ds if not combined_ds else combined_ds.concatenate(ds)
     combined_ds = combined_ds.batch(batch_size, drop_remainder=False)
     # Pad the final batch.
@@ -1347,7 +1719,7 @@ def export_model(estimator, export_dir, vocabulary, sequence_length,
       recent in the model directory.
 
   Returns:
-    None
+    The string path to the exported directory.
   """
 
   def serving_input_fn():
@@ -1381,13 +1753,8 @@ def export_model(estimator, export_dir, vocabulary, sequence_length,
     return tf.estimator.export.ServingInputReceiver(
         features=features, receiver_tensors=inputs)
 
-  tpu_estimator.export_estimator_savedmodel(
-      estimator=estimator,
-      export_dir_base=export_dir,
-      serving_input_receiver_fn=serving_input_fn,
-      as_text=False,
-      checkpoint_path=checkpoint_path,
-  )
+  return estimator.export_saved_model(
+      export_dir, serving_input_fn, checkpoint_path=checkpoint_path)
 
 
 def compute_batch_size(sequence_length,
@@ -1594,12 +1961,13 @@ def run(tpu_job_name,
         tpu, gcp_project, tpu_zone,
         model_dir,
         model_type="bitransformer",
-        vocabulary=gin.REQUIRED,
+        vocabulary=None,
         train_dataset_fn=None,
         eval_dataset_fn=None,
         dataset_split="train",
         autostack=True,
         eval_checkpoint_step=None,
+        export_checkpoint_step=None,
         export_path="",
         mode="train",
         iterations_per_loop=100,
@@ -1639,6 +2007,8 @@ def run(tpu_job_name,
     autostack: boolean, see `get_estimator` docstring for details.
     eval_checkpoint_step: int, list of ints, or None, see `eval_model` doc
       string for details.
+    export_checkpoint_step: int or None, see `export_model` doc string for
+      details.
     export_path: a string, path to export the saved model
     mode: string, train/eval/perplexity_eval/infer
       perplexity_eval computes the perplexity of the dev set.
@@ -1656,7 +2026,10 @@ def run(tpu_job_name,
     mesh_shape: an input to mtf.convert_to_shape()
     mesh_devices: a list of strings, see `get_estimator` docstring.
     layout_rules: an input to mtf.convert_to_layout_rules()
-    learning_rate_schedule: an optional function, see `get_estimator` docstring.
+    learning_rate_schedule: a function which takes the scalar name argument
+      `step` and the numeric argument `total_train_steps` and returns the scalar
+      learning rate.  Alternatively a float.  Alternatively, a list of
+      such factos to be multiplied together.
     optimizer: a class extending optimize.Optimizer, required for training
     predict_fn: an optional function, see `get_estimator` docstring for details.
     variable_filter: a string, see `get_estimator` docstring for details.
@@ -1676,6 +2049,11 @@ def run(tpu_job_name,
   if not isinstance(train_steps, int):
     train_steps = train_steps(batch_size, sequence_length)
 
+  if isinstance(learning_rate_schedule, list):
+    learning_rate_schedule = functools.partial(
+        learning_rate_schedules.product_learning_rate,
+        factors=learning_rate_schedule)
+
   if callable(learning_rate_schedule):
     learning_rate_schedule = functools.partial(
         learning_rate_schedule, total_train_steps=train_steps)
@@ -1758,10 +2136,11 @@ def run(tpu_job_name,
       )
     def _input_fn(params, eval_dataset):
       del params
-      return (eval_dataset.dataset_fn().repeat()
-              .batch(batch_size * (ensemble_inputs or 1),
-                     drop_remainder=True)
-              .prefetch(tf.data.experimental.AUTOTUNE))
+      ds = eval_dataset.dataset_fn().map(_filter_features)
+      ds = transformer_dataset.pad_dataset_with_zeroed_out_examples(ds)
+      ds = (ds.batch(batch_size * (ensemble_inputs or 1), drop_remainder=True)
+            .prefetch(tf.data.experimental.AUTOTUNE))
+      return ds
     checkpoint_paths = get_checkpoint_iterator(eval_checkpoint_step, model_dir)
     for checkpoint_path in checkpoint_paths:
       for eval_dataset in eval_datasets:
@@ -1783,10 +2162,26 @@ def run(tpu_job_name,
   elif mode == "infer":
     infer_model(estimator, vocabulary, sequence_length, batch_size, model_type,
                 model_dir, eval_checkpoint_step)
-
+  elif mode == "score_from_strings":
+    score_from_strings(estimator, vocabulary, model_type, batch_size,
+                       sequence_length, model_dir, eval_checkpoint_step,
+                       dataset_split)
+  elif mode == "score_from_dataset":
+    score_from_dataset(estimator, vocabulary, batch_size, sequence_length,
+                       model_dir, eval_checkpoint_step, dataset_split)
   elif mode == "export":
+    if export_checkpoint_step:
+      checkpoint_path = get_checkpoint_iterator(
+          export_checkpoint_step, model_dir)
+      if isinstance(checkpoint_path, list):
+        checkpoint_path = checkpoint_path[0]
+      else:
+        checkpoint_path = next(checkpoint_path)
+    else:
+      # Use the latest checkpoint in the model directory.
+      checkpoint_path = None
     export_model(estimator, export_path, vocabulary, sequence_length,
-                 batch_size)
+                 batch_size, checkpoint_path)
 
   else:
     raise ValueError(
